/ piper.cpp : Defines the entry point for the console application.
//
#include "stdafx.h"

/****************************************************************************
*                                class CommandLine
****************************************************************************/

class CommandLine {
    public:
       CommandLine() { HTML = FALSE; IsUnicode = FALSE; program = NULL; }
       BOOL HTML;
       BOOL IsUnicode;
       LPTSTR program;
    }; // class CommandLine

/****************************************************************************
*                                class Result
****************************************************************************/

class Result {
    public:
       typedef enum { SUCCESS = 0,
                      NO_PROGRAM,
                      INVALID_ARGUMENT,
                      TWO_COMMANDS,
                      IOCP_FAILED,
                      IOCP_ERROR,
                      THREAD_FAILURE,
                      STDOUT_CREATION_FAILED,
                      STDERR_CREATION_FAILED,
                      CREATEPROCESS_FAILED
       } Type;
};

/****************************************************************************
*                              class SourceFlags
****************************************************************************/

class SourceFlags {
    public:
       typedef enum { None=0, StdOut=1, StdErr=2 } FlagType;
    }; // class SourceFlags

/****************************************************************************
*                              class ThreadParms
****************************************************************************/

class ThreadParms {
    public:
       ThreadParms(HANDLE h, SourceFlags::FlagType f, HANDLE io, BOOL uni) {
          stream = h;
          flags = f;
          iocp = io;
          IsUnicode = uni;
       }
    public:
       HANDLE stream;
       SourceFlags::FlagType flags;
       HANDLE iocp;
       BOOL IsUnicode;
};

/****************************************************************************
*                              class SmartHandle
****************************************************************************/

class SmartHandle {
    public:
       SmartHandle() { handle = NULL; }
       SmartHandle(HANDLE h) { handle = h; }
       virtual ~SmartHandle() { if(handle != NULL) ::CloseHandle(handle); }
    public:
       operator HANDLE() { return handle; }
       operator LPHANDLE() { return & handle; }
       bool operator==(HANDLE h) { return handle == h; }
       SmartHandle & operator=(HANDLE h) { handle = h; return *this; }
    public:
       void Close() { if(handle != NULL) ::CloseHandle(handle); handle = NULL; }
    protected:
       HANDLE handle;
};

/****************************************************************************
*                              class PipeReader
****************************************************************************/

class PipeReader {
    protected:
       static const UINT MAX_BUFFER = 1024;

    public:
       PipeReader(HANDLE str, BOOL uni) { Init(); stream = str; IsUnicode = uni; }
       //****************************************************************
       //                       PipeReader::GetString
       // Result: CString
       //       The most recent string read by the Read method
       // Notes:
       //       This must be done before the next call on Read or the
       //       buffer will be overwritten
       //       Unicode reception is fully supported only in Unicode builds
       //****************************************************************
       CString GetString() { 
           if(IsUnicode) 
               return CString((LPCWSTR)buffer); 
           else 
               return CString((LPCSTR)buffer); }
       //****************************************************************
       //                         PipeReader::Read
       // Result: BOOL
       //       TRUE if there is data in the buffer
       //       FALSE if the pipe has broken
       // Effect:
       //       Reads data from the pipe, and makes it available for
       //       the next GetString call
       // Notes:
       //       If the pipe is being treated as Unicode and an odd number
       //       of bytes has been read, retain the last byte and prepend it
       //       to the next ReadFile buffer
       //****************************************************************
       BOOL Read() {
          if(Offset == 1)
             buffer[0] = reread;

          if(!ReadFile(stream, &buffer[Offset], MAX_BUFFER - (IsUnicode ? sizeof(WCHAR) : sizeof(char)), &bytesRead, NULL))
             return FALSE;

          if(IsUnicode)
             { /* unicode pipe */
              if((Offset + bytesRead) & 1)
                 { /* odd bytes read */
                  Offset = 1; // offset for next read
                  reread = buffer[Offset + bytesRead - 1]; // force reread
                  buffer[Offset + bytesRead - 1] = 0; // remove from current buffer
                  bytesRead--;   // pretend we didn't see it
                 } /* odd bytes read */
              else
                 { /* even bytes read */
                  Offset = 0; // offset for next read
                 } /* even bytes read */

              buffer[Offset + bytesRead] = 0;
              buffer[Offset + bytesRead + 1] = 0; // create Unicode NUL
             } /* unicode pipe */
          else
             { /* ANSI pipe */
              buffer[bytesRead] = '\0';
             } /* ANSI pipe */
         return TRUE;
       } // PipeReader::Read

    protected:
       void Init() { stream = NULL; Offset = 0; IsUnicode = FALSE; }
       BOOL IsUnicode;
       HANDLE stream;
    protected:
       BYTE buffer[MAX_BUFFER];
       DWORD Offset;
       BYTE reread;
       DWORD bytesRead;
}; // class PipeReader

/****************************************************************************
*                                 ErrorString
* Inputs:
*       UINT err: Error code
* Result: CString
*       Result
****************************************************************************/

CString ErrorString(UINT err)
   {
    LPTSTR msg;
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                  NULL,
                  err,
                  0,
                  (LPTSTR) &msg,
                  0,
                  NULL);
    return msg;
   }

/****************************************************************************
*                                   ToHTML
* Inputs:
*       const CString & s:
* Result: CString
*       Modified string with <, > and & replaced with HTML escapes
****************************************************************************/

CString ToHTML(const CString & s)
    {
     CString t = s;
     t.Replace(_T("&"), _T("&amp;")); // this must be the first one
     t.Replace(_T("<"), _T("&lt;"));
     t.Replace(_T(">"), _T("&gt;"));
     return t;
    } // ToHTML

/****************************************************************************
*                               FormatForOutput
* Inputs:
*       CString text: data to show
*       CString & prefix: Prefix for data--leftover partial line from last call
*       ThreadParms * parms: Parameters for the thread
* Result: void
*       
* Effect: 
*       Parses the data into lines.  Retains any partial line (not terminated
*       by a newline sequence) in the Prefix for the next call
* Notes:
*       To force the last partial line out, it must be called with a newline
*       string as text.
*
*       The text is allocated in this thread, on the heap, and must be
*       disposed of by the recipient.
****************************************************************************/

void FormatAndOutput(CString text, CString & prefix, ThreadParms * parms)
    {
     text = prefix + text;

     while(TRUE)
        { /* break into lines */
         int n = text.Find(_T("\r\n"));
         if(n < 0)
            { /* done */
             prefix = text;
             return;
            } /* done */
         CString * s = new CString(text.Left(n));
         
         ::PostQueuedCompletionStatus(parms->iocp, (DWORD)parms->flags, 0, (LPOVERLAPPED)s);
         text = text.Mid(n+2);
        } /* break into lines */
    } // FormatAndOutput

/****************************************************************************
*                                   reader
* Inputs:
*       LPVOID p: (LPVOID)(ThreadParms *) Thread parameters
* Result: UINT
*       irrelevant, 0, always
* Effect: 
*       Parses the data from the stream and emits it as a sequence of lines
****************************************************************************/

UINT __stdcall reader(LPVOID p)
    {
     ThreadParms * parms = (ThreadParms *)p;

     PipeReader pipe(parms->stream, parms->IsUnicode);
     
     CString Prefix;

     while(TRUE)
        { /* processing loop */
         
         if(!pipe.Read())
            { /* failed stream */
             break;
            } /* failed stream */

         FormatAndOutput(pipe.GetString(), Prefix, parms);
        } /* processing loop */

     if(!Prefix.IsEmpty())
        { /* write out last line */
         CString text(_T("\r\n"));
         FormatAndOutput(text, Prefix, parms);
        } /* write out last line */

     ::PostQueuedCompletionStatus(parms->iocp, 0, parms->flags, NULL);
     return 0;
    } // reader

/****************************************************************************
*                                WriteToOutput
* Inputs:
*       const CString & s: String to write
*       SourceFlags::FlagType flag: Flag to indicate source, StdIn or StdOut
*       CommandLine & cmd: Command line options
* Result: void
*       
* Effect: 
*       Writes the string to the output stream
* Notes:
*       
****************************************************************************/

void WriteToOutput(const CString & s, SourceFlags::FlagType flag, CommandLine & cmd)
    {
     if(cmd.HTML)
        { /* HTML */
         CString classname;
         switch(flag)
            { /* decode */
             case SourceFlags::StdOut:
                classname = _T("stdout");
                break;
             case SourceFlags::StdErr:
                classname = _T("stderr");
                break;
            } /* decode */
         _ftprintf(stdout, _T("<p.%s>%s</p>\n"), classname, ToHTML(s));
        } /* HTML */
     else
        { /* console */
         HANDLE console = ::GetStdHandle(STD_OUTPUT_HANDLE);

         CONSOLE_SCREEN_BUFFER_INFO info;
         ::GetConsoleScreenBufferInfo(console, &info);

         switch(flag)
            { /* decode */
             case SourceFlags::StdOut:
                ::SetConsoleTextAttribute(console, FOREGROUND_INTENSITY | FOREGROUND_GREEN);
                break;
             case SourceFlags::StdErr:
                ::SetConsoleTextAttribute(console, FOREGROUND_INTENSITY | FOREGROUND_RED);
                break;
            } /* decode */
         _fputts(s, stdout);
         _fputts(_T("\r\n"), stdout);
         ::SetConsoleTextAttribute(console, info.wAttributes);
        } /* console */
    } // WriteToOutput

/****************************************************************************
*                                   _tmain
* Inputs:
*       int argc: Count of args
*       _TCHAR * argv[]: Array arguments
*                               piper [-u] [-html] command
* Result: int
*       0 for success
*       otherwise failure
* Effect: 
*       Launch the specified program
*       format its output from stdout and stderr into HTML format and send
*       to our stdout
* Notes:
*       The Unicode pipe feature is untested
****************************************************************************/

int _tmain(int argc, _TCHAR* argv[])
   {
    //****************************************************************
    // Argument processing
    //****************************************************************
    if(argc == 1)
       { /* usage */
        CString module;
        LPTSTR p = module.GetBuffer(MAX_PATH);
        ::GetModuleFileName(NULL, p, MAX_PATH);
        module.ReleaseBuffer();
        int n = module.ReverseFind(_T('\\'));
        if(n < 0)
           n = 0;
        module = module.Mid(n + 1);
        n = module.ReverseFind(_T('.'));
        if(n < 0)
           n = module.GetLength();
        module = module.Left(n);
        _ftprintf(stderr, _T("Usage:\n%s [-u] [-html] command\n"), module);
        return Result::INVALID_ARGUMENT;
       } /* usage */

    CommandLine cmd;

    for(int i = 1; i < argc; i++)
       { /* scan args */
        CString arg = argv[i];
        if(arg[0] == _T('-'))
           { /* option */
            if(arg == _T("-u"))
               { /* unicode */
                cmd.IsUnicode = TRUE;
                continue;
               } /* unicode */
            if(arg == _T("-html"))
               { /* html */
                cmd.HTML = TRUE;
                continue;
               } /* html */
            _ftprintf(stderr, _T("Unrecognized option \"%s\"\n"), arg);
            return Result::INVALID_ARGUMENT;
           } /* option */

        if(cmd.program != NULL)
           { /* two files */
            _ftprintf(stderr, _T("Two command directives given:\n  [1] \"%s\"\n  [2]\"%s\"\n"),
                      cmd,
                      arg);
            return Result::TWO_COMMANDS;
           } /* two files */
        cmd.program = argv[i];
       } /* scan args */

    if(cmd.program == NULL)
       { /* no args */
        _ftprintf(stderr, _T("need program to run\n"));
        return Result::NO_PROGRAM;
       } /* no args */

    //****************************************************************
    // Create the I/O Completion Port queue
    //****************************************************************
    SmartHandle iocp = ::CreateIoCompletionPort(INVALID_HANDLE_VALUE, NULL, 0, 0);

    if(iocp == NULL)
       { /* failed iocp */
        DWORD err = ::GetLastError();
        _ftprintf(stderr, _T("CreateIoCompletionPort failed, error %s\n"), ErrorString(err));
        return Result::IOCP_FAILED;
       } /* failed iocp */

    //****************************************************************
    // Create the pipes to route to the child process
    //****************************************************************
    SECURITY_ATTRIBUTES sa = {sizeof(SECURITY_ATTRIBUTES), NULL, TRUE};
    SmartHandle stdout_read;
    SmartHandle stdout_write;
    SmartHandle stderr_read;
    SmartHandle stderr_write;


    static const UINT PIPE_BUFFER_SIZE = 32;

    if(!::CreatePipe((LPHANDLE)stdout_read, (LPHANDLE)stdout_write, &sa, PIPE_BUFFER_SIZE))
       { /* failed stdout */
        DWORD err = ::GetLastError();
        _tprintf(_T("stdout pipe failure: %s\n"), ErrorString(err));
        return Result::STDOUT_CREATION_FAILED;
       } /* failed stdout */

    if(!::CreatePipe((LPHANDLE)stderr_read, (LPHANDLE)stderr_write, &sa, PIPE_BUFFER_SIZE))
       { /* failed stderr */
        DWORD err = ::GetLastError();
        _tprintf(_T("stderr pipe failure: %s\n"), ErrorString(err));
        return Result::STDERR_CREATION_FAILED;
       } /* failed stderr */
    
    //****************************************************************
    // Create the child process
    //****************************************************************
    STARTUPINFO startup = {sizeof(STARTUPINFO)};
    startup.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
    startup.wShowWindow = SW_HIDE;
    startup.hStdOutput = stdout_write;
    startup.hStdError = stderr_write;

    PROCESS_INFORMATION procinfo;

    if(!::CreateProcess(NULL, cmd.program, NULL, NULL, TRUE, CREATE_NEW_CONSOLE, NULL, NULL, &startup, &procinfo))
       { /* failed */
        DWORD err = ::GetLastError();
        _tprintf(_T("CreateProcess failed for \"%s\": %s"), cmd.program, ErrorString(err));
        return Result::CREATEPROCESS_FAILED;
       } /* failed */

    ::CloseHandle(procinfo.hProcess);   // handle will never be needed
    ::CloseHandle(procinfo.hThread);    // handle will never be needed

    stdout_write.Close();               // Close our end of the pipe
    stderr_write.Close();               // Close our end of the pipe

    //****************************************************************
    // Create the threads to handle the pipes
    //****************************************************************
    unsigned id;

    //----------------
    // stdout
    //----------------
    SmartHandle stdoutThread = (HANDLE)_beginthreadex(NULL, 0, reader, new ThreadParms(stdout_read, SourceFlags::StdOut, iocp, cmd.IsUnicode), 0, &id);
    if(stdoutThread == NULL)
       { /* thread create failed */
        DWORD err = ::GetLastError();
        _ftprintf(stderr, _T("Thread creation for stdout failed, error %s\n"), ErrorString(err));
        return Result::THREAD_FAILURE;
       } /* thread create failed */        

    stdoutThread.Close(); // handle will never be used

    //----------------
    // stderr
    //----------------
    SmartHandle stderrThread = (HANDLE)_beginthreadex(NULL, 0, reader, new ThreadParms(stderr_read, SourceFlags::StdErr, iocp, cmd.IsUnicode), 0, &id);
    if(stderrThread == NULL)
       { /* thread create failed */
        DWORD err = ::GetLastError();
        _ftprintf(stderr, _T("Thread creation for stderr failed, error %s\n"), ErrorString(err));
        return Result::THREAD_FAILURE;
       } /* thread create failed */        

    stderrThread.Close(); // handle will never be used

    //****************************************************************
    // Run the loop until both stdout and stderr are broken
    //****************************************************************

    SourceFlags::FlagType broken = SourceFlags::None;

    Result::Type result = Result::SUCCESS;
    
    while(broken != (SourceFlags::StdOut | SourceFlags::StdErr))
       { /* watch pipes */
        OVERLAPPED * ovl;
        DWORD bytesRead;
        ULONG_PTR key;

        //----------------------------------------------------------------
        // bytesRead: the item flag
        //          SourceFlags::StdOut for stdout data
        //          SourceFlags::StdErr for stderr data
        // key: the termination flag
        //          SourceFlags::StdOut when stdout breaks
        //          SourceFlags::StdErr when stderr breaks
        // Note: the <bytesRead, key> pair will either be of the form
        //              <flag_*, 0>   for data notification
        //              <0, flag_*>   for termination notification
        // ovl: (LPOVERLAPPED)(CString *)
        //----------------------------------------------------------------
        BOOL ok = ::GetQueuedCompletionStatus(iocp, &bytesRead, &key, &ovl, INFINITE);

        if(!ok)
           { /* failed */
            DWORD err = ::GetLastError();
            result = Result::IOCP_ERROR;
            _ftprintf(stderr, _T("GetQueuedCompletionStatus failed, error %s\n"), ErrorString(err));
            break;
           } /* failed */

       broken = (SourceFlags::FlagType)(broken | (int)key);
        if(key != 0)
           continue;  // termination notifications contain no data

        CString * s = (CString *)ovl;

        WriteToOutput(*s, (SourceFlags::FlagType)bytesRead, cmd);

        delete s;
       } /* watch pipes */

    //****************************************************************
    // Cleanup
    //****************************************************************
    stdout_read.Close();
    stderr_read.Close();

    return result;
   }  // _tmain

